==> deeno-0.1.tar.gz: deeno-0.1/src/deeno/sql.py <==
"""
SQL query builder.

The parameter `relation` generally refers to a `deeno.db.Relation` instance.

"""
import logging

log = logging.getLogger(__name__)


def _append_where(db, query, params, where):

    if not where:
        return

    if isinstance(where, dict):
        where_clause, where_params = [], []
        for column, value in sorted(where.items()):
            column = db.quote(column)
            where_clause.append('%s = %s' % (column, db.placeholder))
            where_params.append(value)
        where_clause = ' AND '.join(where_clause)
    else:
        where_clause, where_params = where
    query.append('WHERE')
    query.append(where_clause)
    params.extend(where_params)


def _append_returning(db, query, returning):

    if returning:
        if not isinstance(returning, str):
            returning = ', '.join(db.quote(c) for c in returning)
        query.append('RETURNING')
        query.append(returning)


def join(relation, mode, other, on=None, using=None):

    query = ['%s %s %s' % (relation, mode.upper(), other)]
    if on:
        query.append('ON %s' % (on,))
    if using:
        if not isinstance(using, str):
            using = ', '.join(relation.db.quote(c) for c in using)
        query.append('USING (%s)' % (using,))
    return ' '.join(query)


def select(relation, columns='*', where=None, groupby=None, having=None,
        orderby=None, limit=None, offset=None):

    if not isinstance(columns, str):
        columns = ', '.join(relation.db.quote(c) for c in columns)

    query = ["SELECT %s FROM %s" % (columns, relation)]
    params = []

    _append_where(relation.db, query, params, where)

    if groupby:
        if not isinstance(groupby, str):
            groupby = ', '.join(groupby)
        query.append('GROUP BY')
        query.append(groupby)

    if having:
        query.append('HAVING')
        query.append(having)

    if orderby:
        if not isinstance(orderby, str):
            orderby = ', '.join(
                relation.db.quote(c) if isinstance(c, str) else str(c)
                for c in orderby
            )
        query.append('ORDER BY')
        query.append(orderby)

    if offset:
        query.append('OFFSET %s' % (relation.db.placeholder,))
        params.append(offset)

    if limit:
        query.append('LIMIT %s' % (relation.db.placeholder,))
        params.append(limit)

    return ' '.join(query), params


def update(relation, assign, where=None, returning=None):

    if not assign:
        raise ValueError('need something to set')

    query = ["UPDATE %s SET" % (relation,)]
    params = []

    if isinstance(assign, dict):
        assign_clause, assign_params = [], []
        for k, v in sorted(assign.items()):
            k = relation.db.quote(k)
            assign_clause.append('%s = %s' % (k, relation.db.placeholder))
            assign_params.append(v)
        assign_clause = ', '.join(assign_clause)
    else:
        assign_clause, assign_params = assign
    query.append(assign_clause)
    params.extend(assign_params)

    _append_where(relation.db, query, params, where)
    _append_returning(relation.db, query, returning)

    return ' '.join(query), params


def delete(relation, where=None, returning=None):

    query = ["DELETE FROM %s" % (relation,)]
    params = []

    _append_where(relation.db, query, params, where)
    _append_returning(relation.db, query, returning)

    return ' '.join(query), params


def insert(relation, columns=None, rows=None, row=None, returning=False):

    if rows and row:
        raise ValueError('only one of `rows` or `row` may be specified')

    if rows is None and row is None:
        if columns:
            raise ValueError('need some rows')
        rows = [[]]
        columns = ()

    if rows is None:
        rows = [row]

    if columns is None:
        if not isinstance(rows[0], dict):
            raise ValueError('either provide columns ore dictionary rows')
        columns = tuple(sorted(rows[0].keys()))
        if set(tuple(sorted(r.keys())) for r in rows) != set([columns]):
            raise ValueError('dictionary rows must have identical keys')
        rows = [[r[k] for k in columns] for r in rows]
    else:
        if isinstance(columns, str):
            columns = tuple(columns.replace(',', ' ').split())
        if isinstance(rows[0], dict):
            raise ValueError('either provide columns or dict rows, not both')

    nc = len(columns)

    if len(set(len(r) for r in rows)) > 1:
        raise ValueError('rows must have same number of elements')

    if len(rows[0]) != nc:
        raise ValueError('number of columns and row elements differ')

    if not isinstance(columns, str):
        columns = ', '.join(relation.db.quote(c) for c in columns)

    if nc == 0:
        if len(rows) > 1:
            raise ValueError('insert with defaults only works for one row')
        query = ['INSERT INTO %s DEFAULT VALUES' % relation]
        if returning:
            _append_returning(relation.db, query, returning)
        return ' '.join(query), []

    query = ['INSERT INTO %s (%s) VALUES' % (relation, columns)]
    params = []

    placeholders = ['(%s)' % (', '.join([relation.db.placeholder] * nc))]
    placeholders = placeholders * len(rows)
    placeholders = ', '.join(placeholders)
    query.append(placeholders)
    for r in rows:
        params.extend(r)

    _append_returning(relation.db, query, returning)

    return ' '.join(query), params

==> deeno-0.4.tar.gz: deeno-0.4/src/deeno/sql.py <==
"""
SQL query builder.

The parameter `relation` generally refers to a `deeno.db.Relation` instance.

"""
import logging

log = logging.getLogger(__name__)


def _append_where(db, query, params, where):

    if not where:
        return

    if isinstance(where, dict):
        where_clause, where_params = [], []
        for column, value in sorted(where.items()):
            column = db.quote(column)
            where_clause.append('%s = %s' % (column, db.placeholder))
            where_params.append(value)
        where_clause = ' AND '.join(where_clause)
    else:
        where_clause, where_params = where
    query.append('WHERE')
    query.append(where_clause)
    params.extend(where_params)


def _append_returning(db, query, returning):

    if returning:
        if not isinstance(returning, str):
            returning = ', '.join(db.quote(c) for c in returning)
        query.append('RETURNING')
        query.append(returning)


def join(relation, mode, other, on=None, using=None):

    query = ['%s %s %s' % (relation, mode.upper(), other)]
    if on:
        query.append('ON %s' % (on,))
    if using:
        if not isinstance(using, str):
            using = ', '.join(relation.db.quote(c) for c in using)
        query.append('USING (%s)' % (using,))
    return ' '.join(query)


def select(relation, columns='*', where=None, groupby=None, having=None,
        orderby=None, limit=None, offset=None):

    if not isinstance(columns, str):
        columns = ', '.join(relation.db.quote(c) for c in columns)

    query = ["SELECT %s FROM %s" % (columns, relation)]
    params = []

    _append_where(relation.db, query, params, where)

    if groupby:
        if not isinstance(groupby, str):
            groupby = ', '.join(groupby)
        query.append('GROUP BY')
        query.append(groupby)

    if having:
        query.append('HAVING')
        query.append(having)

    if orderby:
        if not isinstance(orderby, str):
            orderby = ', '.join(
                relation.db.quote(c) if isinstance(c, str) else str(c)
                for c in orderby
            )
        query.append('ORDER BY')
        query.append(orderby)

    if offset:
        query.append('OFFSET %s' % (relation.db.placeholder,))
        params.append(offset)

    if limit:
        query.append('LIMIT %s' % (relation.db.placeholder,))
        params.append(limit)

    return ' '.join(query), params


def update(relation, assign, where=None, returning=None):

    if not assign:
        raise ValueError('need something to set')

    query = ["UPDATE %s SET" % (relation,)]
    params = []

    if isinstance(assign, dict):
        assign_clause, assign_params = [], []
        for k, v in sorted(assign.items()):
            k = relation.db.quote(k)
            assign_clause.append('%s = %s' % (k, relation.db.placeholder))
            assign_params.append(v)
        assign_clause = ', '.join(assign_clause)
    else:
        assign_clause, assign_params = assign
    query.append(assign_clause)
    params.extend(assign_params)

    _append_where(relation.db, query, params, where)
    _append_returning(relation.db, query, returning)

    return ' '.join(query), params


def delete(relation, where=None, returning=None):

    query = ["DELETE FROM %s" % (relation,)]
    params = []

    _append_where(relation.db, query, params, where)
    _append_returning(relation.db, query, returning)

    return ' '.join(query), params


def insert(relation, columns=None, rows=None, row=None, returning=False):

    if rows and row:
        raise ValueError('only one of `rows` or `row` may be specified')

    if rows is None and row is None:
        if columns:
            raise ValueError('need some rows')
        rows = [[]]
        columns = ()

    if rows is None:
        rows = [row]

    if columns is None:
        if not isinstance(rows[0], dict):
            raise ValueError('either provide columns or dictionary rows')
        columns = tuple(sorted(rows[0].keys()))
        if set(tuple(sorted(r.keys())) for r in rows) != set([columns]):
            raise ValueError('dictionary rows must have identical keys')
        rows = [[r[k] for k in columns] for r in rows]
    else:
        if isinstance(columns, str):
            columns = tuple(columns.replace(',', ' ').split())
        if isinstance(rows[0], dict):
            raise ValueError('either provide columns or dict rows, not both')

    nc = len(columns)

    if len(set(len(r) for r in rows)) > 1:
        raise ValueError('rows must have same number of elements')

    if len(rows[0]) != nc:
        raise ValueError('number of columns and row elements differ')

    if not isinstance(columns, str):
        columns = ', '.join(relation.db.quote(c) for c in columns)

    if nc == 0:
        if len(rows) > 1:
            raise ValueError('insert with defaults only works for one row')
        query = ['INSERT INTO %s DEFAULT VALUES' % relation]
        if returning:
            _append_returning(relation.db, query, returning)
        return ' '.join(query), []

    query = ['INSERT INTO %s (%s) VALUES' % (relation, columns)]
    params = []

    placeholders = ['(%s)' % (', '.join([relation.db.placeholder] * nc))]
    placeholders = placeholders * len(rows)
    placeholders = ', '.join(placeholders)
    query.append(placeholders)
    for r in rows:
        params.extend(r)

    _append_returning(relation.db, query, returning)

    return ' '.join(query), params
